//
// motionpainted.sl surface shader (RenderMan Interface Specification)
//
// Kevin Egan
// ktegan@cs.columbia.edu
//
// Frequency Analysis and Sheared Reconstruction for Rendering Motion Blur
// SIGGRAPH 2009
//
// 
// This surface shader computes a basic shading model with textured surfaces,
//   multiple lights, and phong reflections.  This can output color as
//   well as speed and spatial frequency of the signal.  This shader
//   computes frequencies for one shading model, the theory can be used
//   to compute frequency bounds for many other types of shading models.
//
// This shader takes multiple discrete light sources instead of a single
//   environment as is discussed in the paper.  Evaluation of each light
//   source follows the theory and the algorithm laid out in the paper.
//   The only thing we have to add is a final step that sums each
//   lights contribution and computes the frequency bound for the sum of
//   light signals.
//
//  Because we have direct access to shader derivatives
//   some things (like the change in the reflection vector) can be
//   expressed directly, making the computations slightly different than
//   the paper.  To have this shader work properly for motion
//   blurred shadows and reflections you need to shade at every sample
//   (no caching of data in micropolygons).  PRMan's multi-segment motion
//   blur is clunky for this purpose.  Pixie and 3Delight both have ray
//   tracing modes that support this.  We used Pixie.
//
// The code has some constructs to work around current Pixie issues.
//   We have found that Pixie has issues computing dPdu and other
//   derivatives for bezier surfaces.  For this reason, for the teapot
//   scene, we actually used PRMan for intial sampling, and then
//   used Pixie for final sampling when we needed to have more
//   precise control over the space-time sample placements.




// We assume that coordinate systems for each light are declared
//   in the .rib file, named "lightcoord1", "lightcoord2" etc.
//   We also assume the center of the light is at (0, 0, 0) in this
//   coord system.
// 
// Output:
//   Ci - Out color
//   OutMotion - 2D image space normalized direction
//   OutBounds - a_min, a_max, and omega_x_max
// 
// This shader is for mid to far illumination with discrete light sources.
//   Because we separate the signals per light the final signal is a sum
//   of signals.  For each light we have a shadow, BRDF and surface
//   signal that are multiplied per light.  After that all light signals
//   are summed.
//
// For now we simply use a cone of directions to fake area lights (only
//   used for shadows).  Future shader versions will ray trace towards light
//   geometry.  The shad_freq parameters are the frequency of the area
//   light signal (the occluding frequency is assumed to be infinite).
//   Reflections are treated as a final light source, and the
//   refl_freq parameter is the frequency of the reflected scene.
//
surface motionpainted(
    uniform float Ka = 1;
    uniform float Kd = .5;
    uniform float Ks = .5;
    uniform float do_freq_calc = 1;
    uniform float do_shading = 1;
    uniform string texturename = "";
    uniform float surface_freq = 1.0;
    uniform float phong_exp = -1.0;         // if < 0 then diffuse + mirror
    uniform float brdf_freq = -1.0;         // if < 0 calc automatically
    uniform float use_light1 = 1;
    uniform float use_light2 = 0;
    uniform float use_light3 = 0;
    uniform float shad1_cone = 0.0;         // radians, 0 == point light
    uniform float shad2_cone = 0.0;
    uniform float shad3_cone = 0.0;
    uniform float shad1_freq = 0.8;         // wavelengths per radian
    uniform float shad2_freq = 0.8;
    uniform float shad3_freq = 0.8;
    uniform float shad1_num_rays = 1;
    uniform float shad2_num_rays = 1;
    uniform float shad3_num_rays = 1;
    uniform float has_reflection = 0.0;
    uniform float refl_freq = 30.0;         // wavelengths per radian
    uniform float refl_cone = 0.0;          // radians, 0 == perfect mirror
    uniform float refl_num_rays = 1;
    output varying vector OutMotion = vector(0, 0, 0);
    output varying vector OutBounds = vector(0, 0, 0);
    )
{
    // ********************************************************
    // DECLARE CONSTANTS
    // ********************************************************

    // Pixie doesn't currently work with #define
    //
    // Per light there are shading and shadow signals, separate from that we
    //   have the surface signal.  Additionally an optional glossy
    //   reflection which is considered to be an extra light source.
    //
    // For 3 lights plus a glossy reflection we would use the the
    //     following signal indices in a 9 element array.  Note that
    //     that for each discrete light source we consider different
    //     parts of the hemispherical BRDF and shadow signals. 
    //   0: surface texture, 1: BRDF1, 2: SHADOW1, 3: BRDF2, 4: SHADOW2
    //   5: BRDF3, 6: SHADOW3, 7: <not used>, 8: REFLECTION
    float MAX_LIGHTS = 4;       // includes reflection 
    float NONLIGHT_SIGNALS = 1; // num signals not associated with a light
    float LIGHT_SIGNALS = 2;    // num signals per light
    float SURFACE_INDEX = 0;    // index of surface signal
    float BRDF_OFFSET = 0;      // offset (per light) of BRDF signal
    float SHADOW_OFFSET = 1;    // offset (per light) of shadow signal

    float MIN_LENGTH = 0.0001;  // try to prevent dividing by 0
    float MAX_ANGLE_COS = 0.940;// maximum signal direction variation
                                // 0.940 = cos(20 * PI / 180)
    float COLOR_CUTOFF = 0.001; // don't count frequency contributions
                                // below cutoff 


    // ********************************************************
    // INITIALIZE LIGHT DATASTRUCTURES
    // ********************************************************

    point light_pos[4];     // size = MAX_LIGHTS (Pixie workaround)
    vector light_dir[4];    //   can't use #define
    float light_freq[4];
    float light_dist[4];
    float light_used[4];
    color light_contrib[4];
    float shad_cone[3];
    float shad_freq[3];
    float shad_numrays[3];


    // we want to set a limit on ray tracing
    float raydepth;
    rayinfo("depth", raydepth);

    string rayname;
    rayinfo("label", rayname);

    if (rayname == "noshade") {
        // if this is a shadow ray or a frequency query ray
        // then don't do any recursive ray tracing
        raydepth = 10;
    }

    float eff_do_freq_calc;
    float eff_do_shading;
    if (raydepth < 1) {
        // primary ray
        eff_do_freq_calc = 1;
        eff_do_shading = 1;
    } else if (raydepth < 3) {
        // second and third bounce
        // reflections and shadows are also set to 1 ray later
        eff_do_freq_calc = 0;
        eff_do_shading = 1;
    } else {
        eff_do_freq_calc = 0;
        eff_do_shading = 0;
    }


    // We assume that coordinate systems for each light are declared
    // in the .rib file, named "lightcoord1", "lightcoord2" etc.
    light_pos[0] = transform("current", point "lightcoord1" (0, 0, 0));
    light_pos[1] = transform("current", point "lightcoord2" (0, 0, 0));
    light_pos[2] = transform("current", point "lightcoord3" (0, 0, 0));
    light_pos[3] = 0;                   // reserved for reflection
    light_used[0] = use_light1;
    light_used[1] = use_light2;
    light_used[2] = use_light3;
    light_used[3] = 0;                  // reserved for reflection
    light_freq[0] = 1e4;                // BRDF shading uses point lights
    light_freq[1] = 1e4;                //   for now which have high frequency
    light_freq[2] = 1e4;
    light_freq[3] = 0;                  // reserved for reflection

    shad_cone[0] = shad1_cone;
    shad_cone[1] = shad2_cone;
    shad_cone[2] = shad3_cone;
    shad_freq[0] = shad1_freq;
    shad_freq[1] = shad2_freq;
    shad_freq[2] = shad3_freq;
    shad_numrays[0] = shad1_num_rays;
    shad_numrays[1] = shad2_num_rays;
    shad_numrays[2] = shad3_num_rays;
    if (raydepth > 0) {
        // don't do soft shadows for secondary rays
        shad_cone[0] = 0;
        shad_cone[1] = 0;
        shad_cone[2] = 0;
        shad_freq[0] = 1e4;
        shad_freq[1] = 1e4;
        shad_freq[2] = 1e4;
        shad_numrays[0] = 1;
        shad_numrays[1] = 1;
        shad_numrays[2] = 1;
    }


    // compute BRDF bandlimit with possible phong lobe
    float eff_brdf_freq = brdf_freq;
    if (eff_brdf_freq < 0) {
        if (phong_exp > 0.2) {
            // from "A Signal-Processing Framework for Inverse Rendering"
            // Siggraph 2001.  This is in sterradians, probably needs some
            // modification.
            eff_brdf_freq = sqrt(6 * phong_exp);
        } else {
            eff_brdf_freq = 0;  // set cosine lobe of BRDF has having low frequency
        }
    }

    // compute BRDF bandlimit for reflections with either phong
    // or mirror reflections
    float eff_refl_num_rays = refl_num_rays;
    float eff_refl_cone = refl_cone;
    float eff_refl_brdf_freq = eff_brdf_freq;
    if (has_reflection > 0 && (phong_exp < 0 || refl_num_rays == 1 ||
            refl_cone == 0.0 || raydepth > 0)) {
        // mirror brdf
        eff_refl_num_rays = 1;
        eff_refl_cone = 0;
        eff_refl_brdf_freq = 1e4;
    }


    float num_lights = MAX_LIGHTS;

    float lightindex;
    for (lightindex = 0; lightindex < MAX_LIGHTS; lightindex += 1) {
        if (light_used[lightindex] == 0) {
            num_lights = min(num_lights, lightindex);
            light_dir[lightindex] = vector(0, -1, 0);
            light_dist[lightindex] = -1;
        } else {
            light_dir[lightindex] = light_pos[lightindex] - P;
            light_dist[lightindex] = length(light_dir[lightindex]);
            light_dir[lightindex] = normalize(light_dir[lightindex]);
        }
        light_contrib[lightindex] = 0;
    }

    float num_lights_wrefl = num_lights;
    if (has_reflection > 0) {
        num_lights_wrefl = num_lights + 1;
        light_freq[num_lights] = refl_freq;
    }


    float num_signals = num_lights_wrefl * LIGHT_SIGNALS + NONLIGHT_SIGNALS;
    if (eff_do_freq_calc == 0) {
        num_signals = 0;
    }


    // ********************************************************
    // COMPUTE UNSHADOWED SHADING PER LIGHT
    // FREQUENCIES COMPUTED LATER
    // ********************************************************
    color Ct;


    if(texturename != "")
        Ct = color texture(texturename);
    else
        Ct = 1;

    normal normN = normalize(N);
    normal Nf = normalize(faceforward (normN, I));
    vector V = -normalize(I);
    Oi = Os;
    

    // currently no ambient contribution
    // Ci = Os * (Ct * (Ka * ambient()));
    Ci = 0;

    // loop over each incoming light source, match light indices
    // to lights, and store unshadowed contribution in light_contrib
    illuminance( P, Nf, PI/2 ) {
        vector normL = normalize(L);

        // diffuse shading per light
        color diffuse_contrib = Kd * Cl * (normL . Nf);
        color phong_contrib = 0;

        // phong shading per light
        if (phong_exp >= 0) {
            vector normI = normalize(I);
            vector reflectDir = reflect(normI, Nf);
            phong_contrib = Ks * Cl * pow(max(0, normL . reflectDir), phong_exp);
        }

        // find the light in the light_dir array that matches the L direction
        float bestmatch = -1;
        float indexmatch = -1;
        for (lightindex = 0; lightindex < num_lights; lightindex += 1) {
            float cosangle = light_dir[lightindex] . normL;
            if (cosangle > bestmatch) {
                bestmatch = cosangle;
                indexmatch = lightindex;
            }
        }
        if (bestmatch > 0.99) {
            light_contrib[indexmatch] = diffuse_contrib + phong_contrib;
            if (phong_contrib[0] < COLOR_CUTOFF &&
                    phong_contrib[1] < COLOR_CUTOFF &&
                    phong_contrib[2] < COLOR_CUTOFF) {
                light_freq[indexmatch] = 0;
            }
        }
    }



    // ********************************************************
    // INITIALIZE FREQUENCY DATASTRUCTURES
    // ********************************************************

    // effective velocity 'a' and omega_x_max frequency for each signal
    vector signal_vel[9];   // size = MAX_SIGNALS
    float signal_freq[9];   // size = MAX_SIGNALS

    float signalindex;
    for (signalindex = 0; signalindex < num_signals; signalindex += 1) {
        signal_vel[signalindex] = vector(0, 0, 0);
        signal_freq[signalindex] = 0;
    }


    // ********************************************************
    // COMPUTE BRDF SHADING FREQUENCY AND REFLECTION COLOR
    // STORE FREQUENCY INFO AND ADD REFLECTION COLOR TO FINAL COLOR
    // ********************************************************

    color reflect_color = 0;

    
    // Light source color contributions have already been computed.
    // Reflection color contributions still need to be computed.
    //
    // TODO some of these computations do not need to be recalculated
    // per light.

    for (lightindex = 0; lightindex < num_lights_wrefl; lightindex += 1) {
        vector sourceMotion = -1;
        vector sourceDir = -1;
        float sourceDist = -1;

        if (lightindex < num_lights) {
            // normal light source
            sourceDir = light_dir[lightindex];
            sourceDist = light_dist[lightindex];

            // TODO have light imposter geometry so that we can ray trace
            // and sample the light motion.
            sourceMotion = vector(0, 0, 0);

            color cur_contrib = light_contrib[lightindex];
        } else {
            // treat reflection as a light source

            vector normI = normalize(I);
            sourceDir = reflect(normI, Nf);

            color surfaceColor = -1;


            // cast multiple rays for computing glossy reflection color
            if (eff_do_shading > 0 && eff_refl_num_rays > 0) {
                float cone = eff_refl_cone;
                float num_rays = eff_refl_num_rays;
                if (phong_exp < 0) {
                    cone = 0;
                    num_rays = 1;
                }

                vector raydir = -1;

                // ray trace
                gather("", P, sourceDir, cone, num_rays,
                        "surface:Ci", surfaceColor,
                        "ray:direction", raydir) {
                    vector refl_dir = normalize(raydir);
                    if (phong_exp < 0) {
                        // mirror reflection
                        reflect_color += surfaceColor;
                    } else {
                        // phong reflection shading
                        float brdf = pow(max(0, (refl_dir . sourceDir)), phong_exp);
                        reflect_color += surfaceColor * brdf;
                    }
                } else {
                    // no reflection, ray trace missed
                }

                reflect_color /= num_rays;

                // this is a reflection, add in the color now
                // (no shadow for a reflection)
                Ci += Ks * reflect_color;
            }

            // cast one ray for computing reflection frequency
            if (eff_do_freq_calc > 0) {
                // ray trace
                gather("", P, sourceDir, 0, 1,
                        "label", "noshade",
                        "surface:dPdtime", sourceMotion,
                        "ray:length", sourceDist) {
                    // sourceDist has been set
                } else {
                    // no reflection, ray trace missed
                    sourceDist = -1;
                    sourceMotion = -1;
                    surfaceColor = -1;
                }
            }
        }


        if (eff_do_freq_calc > 0 && sourceDist > -1 &&
            (lightindex == num_lights || light_freq[lightindex] > 0) &&
            (lightindex == num_lights || eff_brdf_freq > 0)) {
            // We want to find the effective velocity for a moving
            // reflection.  This means the image space direction we need
            // to follow to track the same point on the reflected
            // geometry.  This becomes trickier in 3D because we
            // follow a line along the surface the reflection vector
            // may rotate out of plane.  Luckily shader derivatives
            // can give us many of the values directly.
            //
            // The paper has gamma = Beta * kappa' + alpha.  In
            // the shader it is more convenient to directly compute
            // the Beta * kappa' portion through shader derivatives,
            // ie the change of the reflection vector dRdtime.
            //
            // Through linearization we express all angles as tangent
            // vectors 1 unit from the surface.
            vector unit_dPdu = normalize(dPdu);
            vector unit_dPdv = normalize(dPdv);


            // Project dPdtime onto tangent surface as seen from view V
            // get vectors tangent to view V
            vector dPdtime_tan = dPdtime - V * (dPdtime . V);
            vector dPdu_tan = dPdu - V * (dPdu . V);
            vector dPdv_tan = dPdv - V * (dPdv . V);

            // take dot product in plane tanget to view
            // dPdtime_u = cos(dPdtime_tan, dPdu_tan) *
            //    length(dPdtime_tan) / length(dPdu_tan)
            float dPdtime_u = (dPdtime_tan . dPdu_tan) / max(MIN_LENGTH, dPdu_tan . dPdu_tan);
            float dPdtime_v = (dPdtime_tan . dPdv_tan) / max(MIN_LENGTH, dPdv_tan . dPdv_tan);

            // scale so that we measure distance along surface tangent
            float uScale = min(5, length(dPdu) / length(dPdu_tan));
            float vScale = min(5, length(dPdv) / length(dPdv_tan));

            // Note, if the surface travels along the direction of
            // dPdu then at a fixed pixel u will decrease with time.
            // -dudtime * dPdu = projection along view V of dPdtime onto dPdu
            float dudtime = -dPdtime_u * uScale;
            float dvdtime = -dPdtime_v * vScale;


            // use shader derivatives to calculate change in reflection
            // vector along dPdu and dPdv axes.  This information is
            // similar to the curvature in the flatland case, since
            // it gives us change in angle relative to change along
            // the surface.  Make sure vectors are pependicular to
            // sourceDir.
            vector dRdu = Du(sourceDir);
            vector dRdv = Dv(sourceDir);

            // dRdtime is the equivalent of Beta * kappa' in the paper.
            // Note that this is actually more accurate, because it should
            // also include change in viewing direction for a perspective
            // camera.
            vector dRdtime = dRdu * dudtime + dRdv * dvdtime;
            vector unit_dRdtime = normalize(dRdtime);

            // compute alpha, angular movement of light,
            // linearization of angular motion = tangent motion / distance
            vector surf_tangent_motion = sourceMotion -
                sourceDir * (sourceMotion . sourceDir);
            vector alpha = surf_tangent_motion / max(MIN_LENGTH, sourceDist);

            // subtract these to get relative change of reflection
            // vector relative to motion of reflector
            vector gamma = dRdtime - alpha;
            vector unit_gamma = normalize(gamma);

            // index into the last light reserved for reflections
            float brdfindex = NONLIGHT_SIGNALS +
                lightindex * LIGHT_SIGNALS + BRDF_OFFSET;


            // Check if this is a moving signal
            if (length(gamma) > 0) {

                // If we have near and mid-range illumination we
                // need to take that into account.  This is an extra
                // term that we did not have room to discuss in the
                // paper.
                //
                // a = source speed
                // b = surface speed
                // c = curvature
                // d = distance to illumination
                //
                // distant illumination:
                // a = b * c * d
                //
                // near and mid-range illumination, factor in surface
                //   movement b:
                // a - b = b * c * d
                // a = b * c * d + b
                // a / (c * d + 1) = b
                // b = (a / d) / (c + 1 / d)
                //
                // (a / d) is the angular speed, and (c + 1 / d) is the
                // curvature.  When d is large this will converge 
                // to simply c as expected.
                //
                // Add this term to the dRdu and dRdv curvature vectors
                // (accounting for dRdu being measured with respect to
                // u instead of space).  This will make reflections of
                // near sources move more slowly.
                dRdu = dRdu + normalize(dRdu) * length(dPdu) / sourceDist;
                dRdv = dRdv + normalize(dRdv) * length(dPdv) / sourceDist;

                // Now we need to compute what direction along the surface
                // that will cause the reflection vector R to change in the
                // same direction and speed as gamma.  In flatland this was
                // simply dividing by kappa' the curvature.  In 3D
                // it is more complicated because for each axis dPdu and
                // dPdv we have a vector dRdu and dRdv that describes
                // the change in R.
                //
                // First we project gamma using dRdu and dRdv as bases. 
                // Because we are looking at changes in R instead of
                // in N we don't need the factor of 2 that the paper has.
                //
                // Our goal in this step is to compute signal speed measured
                // by change in uv coordinates per time.
                //
                // dRdu and dRdv may not be orthogonal.  We find
                // dsignalu_dtime and dsignalv_dtime that express gamma
                // in terms of dRdu and dRdv.  To do this we need to
                // solve a small 2x2 system of equations in the tangent
                // plane.  In the tangent plane we use a coordinate system
                // where dRdu is one basis vector and perpdRdu, a perpendicular
                // vector of the same length, is the other
                // basis vector.  The variables: a matrix expressing dRdu
                // and dRdv bases in tangent plane, b contains the dRdu
                // and dRdv coefficients that we want to compute,
                // c is gamma in tangent plane.
                //
                // a = | 1   dRdv . dRdu_unit / length(dRdu)         |  
                //     | 0   dRdv . perpdRdu_unit / length(perpdRdu) | 
                //
                // b = | coeff1 |
                //     | coeff2 |       
                //
                // c = | gamma . dRdu_unit / length(dRdu)         |
                //     | gamma . perpdRdu_unit / length(perpdRdu) |
                // 
                // d = a^-1 = |1 x|^{-1} = 1/(y) * | y  -x|
                //            |0 y|                | 0   1|
                //
                // a * b = c
                // b = d * c
                //
                // dsignalu_dtime = b[0] = d[0][0] * c[0] + d[0][1] * c[1]
                // dsignalv_dtime = b[1] = d[1][0] * c[0] + d[1][1] * c[1]
                vector perpdRdu = Nf ^ dRdu;
                perpdRdu *= length(dRdu) / length(perpdRdu);
                float x = (dRdv . dRdu) / (dRdu . dRdu);
                float y = (dRdv . perpdRdu) / (perpdRdu . perpdRdu);
                vector c;
                c[0] = (gamma . dRdu) / (dRdu . dRdu);
                c[1] = (gamma . perpdRdu) / (perpdRdu . perpdRdu);
                float dsignalu_dtime = (1.0 / y) * (y * c[0] + -x * c[1]);
                float dsignalv_dtime = (1.0 / y) * (0 * c[0] + 1 * c[1]);


                // transform this into world coordinates, then image pixel
                // coordinates
                vector signal_velocity = dPdu * dsignalu_dtime + dPdv * dsignalv_dtime;
                vector screen_velocity = transform("raster", P + signal_velocity) - transform("raster", P);
                signal_vel[brdfindex] = screen_velocity;
            }


            // Signal frequency is the minimum frequency of BRDF and light
            // source.  For the BRDF bandlimit Eqn. 37 says to multiply
            // the angular frequency bandlimit by the curvature kappa'.
            // In the 3D case we have a more general mapping.
            //
            // To measure the frequency bandlimit we need to find the
            // maximum compression of the signal as it transforms from
            // angular frequencies to image space.  One method is to
            // calculate the transformation matrix from the tangent
            // space of the surface (our linearized angular space)
            // to raster coordinates.  Then calculate the smallest
            // eigenvalue of that transformation.  Instead we currently
            // approximate the scaling by measuring how 7 test vectors
            // are scaled.  The following uses 4 uniformly sampled
            // directions across a semi-circle, and 3 test vectors
            // (direction of surface slope from view, direction
            // perpendicular to dRdu and dRdv).  The 3 test vectors are
            // chosen account for cases where one factor is dominant (slope of
            // surface and one small dRdu or dRdv basis vector respectively).
            // 
            // Unit length tangent vector corresponds to one radian in
            // linearized tangent domain.
            float NUM_TEST_VECTORS = 7;
            vector testvecs[7];

            vector dPdu_perp = Nf ^ dPdu;
            testvecs[0] = normalize(dPdu);
            testvecs[1] = normalize(dPdu + dPdu_perp);
            testvecs[2] = normalize(dPdu_perp);
            testvecs[3] = normalize(dPdu_perp - dPdu);
            vector temp = Nf ^ V;
            vector Ntangent = Nf ^ temp;
            if (length(Ntangent) > MIN_LENGTH) {
                testvecs[4] = normalize(Ntangent);
            } else {
                testvecs[4] = dPdu;
            }
            testvecs[5] = normalize(Nf ^ dRdu);
            testvecs[6] = normalize(Nf ^ dRdv);

            float minDist = 1e4;
            float testindex;
            for (testindex = 0; testindex < NUM_TEST_VECTORS; testindex = testindex + 1) {
                vector curvec = testvecs[testindex];
                float dudAngle = (curvec . dRdu) / max(MIN_LENGTH, (dRdu . dRdu));
                float dvdAngle = (curvec . dRdv) / max(MIN_LENGTH, (dRdv . dRdv));

                // transform this into world coordinates, then image pixel
                // coordinates
                vector dPdAngle = dPdu * dudAngle + dPdv * dvdAngle;
                vector dPixdAngle = transform("raster", P + dPdAngle) - transform("raster", P);

                float screenDist = length(dPixdAngle);
                minDist = min(minDist, screenDist);
            }


            // we transformed a tangent unit vector (~ 1 radian) into
            // minDist pixels
            // effective curvature = (1 radian) / (dist pixels)
            float kappa = 1.0 / max(MIN_LENGTH, minDist);
            float freq = eff_brdf_freq;
            if (lightindex == num_lights) {
                // use frequency of surrounding scene
                freq = refl_freq;
            }
            signal_freq[brdfindex] =
                min(light_freq[lightindex] * kappa, freq * kappa);

        } else {
            // don't calc frequencies or there is no reflection,
            // signal_vel and signal_freq are already zero
            // so we don't have to do anything
        }

    }  // for each light



    // ********************************************************
    // COMPUTE SHADOW AND SHADOW FREQUENCY (PER LIGHT)
    // ADD MODULATED RESULTS FROM UNSHADOWED SHADING TO FINAL COLOR
    // ********************************************************

    for (lightindex = 0; lightindex < num_lights; lightindex += 1) {
        point lightP = light_pos[lightindex];

        // TODO have light imposter geometry so that we can ray trace
        // and sample the light motion.
        vector lightMotion = vector(0, 0, 0);

        vector lightDir = (lightP - P);
        float lightDist = length(lightDir);
        lightDir = lightDir / lightDist;

        if (Nf . lightDir < -0.1) {
            // back facing surface doesn't really have shadow
            // do nothing
        } else {

            vector occluderMotion = -1;
            float occlusion = 0.0;

            if (eff_do_shading > 0 && shad_numrays[lightindex] > 0) {
                // trace one or more rays, multiple rays gives soft shadow from
                // fake area light (cone of directions)
                gather("", P, lightDir, shad_cone[lightindex], shad_numrays[lightindex],
                        "label", "noshade",
                        "maxdist", lightDist) {
                    // in shadow
                    occlusion += 1.0;
                } else {
                    // no shadow
                }
                occlusion /= shad_numrays[lightindex];

                // add light contribution modulated by shadow
                Ci += Ct * light_contrib[lightindex] * (1.0 - occlusion);
            }

            vector curshad;
            float shadDist = -1;
            if (eff_do_freq_calc > 0) {
                // trace one ray to query occluder info
                gather("", P, lightDir, 0, 1,
                        "maxdist", lightDist,
                        "label", "noshade",
                        "surface:dPdtime", occluderMotion,
                        "ray:length", shadDist) {
                    // shadDist has been set
                    lightDir = normalize(lightDir);
                } else {
                    // no shadow, reset shadDist
                    shadDist = -1;
                }
            }

            if (shadDist >= 0) {
                // This follows Sec. 3.3 of the paper, and also accounts
                // for angle between the receiver and the view. Instead
                // of explicitly computing the cosine term the code
                // below projects speeds into a plane tangent to the
                // direction of mu(x).  The shad1_freq and similar
                // parameters pass in the frequency of the light source
                // being shadowed.  We assume the light source will have
                // lower frequencies than the occluder.
                //
                // As noted in the paper it is easiest to do the
                // calculations in a reference frame where the receiver
                // does not move.  We subtract dPdtime from shadow and
                // occluder velocities.
                occluderMotion = occluderMotion - dPdtime;
                lightMotion = lightMotion - dPdtime;

                // occluder motion perpendicular to shadow direction
                // still need to divide by occluder distance to get
                // relative angular motion, Beta * nu
                // tanOccMotion = Beta * nu * D
                vector tanOccMotion = occluderMotion - lightDir * (occluderMotion . lightDir);
                // light motion perpendicular to shadow direction
                // tanLightMotion = alpha * lightDistance
                vector tanLightMotion = lightMotion - lightDir * (lightMotion . lightDir);

                // This is the relative angular speed.  Because our
                // light source is not infinitely far away we use
                // a slightly modified geometric relationship,
                // the (shadDist) / (lightDist - shadDist) term.
                // tangentMotion = alpha + Beta * nu
                vector tangentMotion = (tanOccMotion - tanLightMotion) * (shadDist / (lightDist - shadDist)) + tanOccMotion;

                // Take into account angle between the receiver and the
                // view.
                float nDotMotion = Nf . normalize(tangentMotion);
                float sinAngle = max(0.2, sqrt(1.0 - nDotMotion * nDotMotion));
                float surfaceScale = length(tangentMotion) / sinAngle;
                vector surfaceMotion = normalize(tangentMotion - Nf * (tangentMotion . Nf)) * surfaceScale;

                // we subtracted dPdtime before, put this back to
                // get into regular coordinate frame
                surfaceMotion = surfaceMotion + dPdtime;

                // transform shadow movement to screen coordinates
                point EndP = P + surfaceMotion;
                vector screen_velocity = transform("raster", EndP) - transform("raster", P);

                // compute frequency information
                float nu = (lightDir . Nf) / lightDist;
                float shadowindex = NONLIGHT_SIGNALS +
                    lightindex * LIGHT_SIGNALS + SHADOW_OFFSET;
                signal_vel[shadowindex] = screen_velocity;
                signal_freq[shadowindex] = shad_freq[lightindex] * nu;
            }
        }
    }



    // ********************************************************
    // COMPUTE FREQUENCY FOR SURFACE TEXTURE
    // ********************************************************

    point EndP = P + dPdtime;
    
    // dPdtime vector projected into screen coordinates
    signal_vel[SURFACE_INDEX] = transform("raster", EndP) - transform("raster", P);
    signal_freq[SURFACE_INDEX] = surface_freq;



    // ********************************************************
    // PROCESS FREQUENCY INFORMATION
    // ********************************************************

    float signal_a[9];       // size = MAX_SIGNALS
    vector signal_norm[9];   // size = MAX_SIGNALS

    vector avg_vel = vector(0, 0, 0);
    float avg_count = 0;

    // separate directions and magnitudes, compute avg velocities
    for (signalindex = 0; signalindex < num_signals; signalindex += 1) {
        if (signal_freq[signalindex] == 0) {
            // mark as no signal
            signal_vel[signalindex] = vector(0, 0, 0);
            signal_norm[signalindex] = vector(0, 0, 0);
            signal_a[signalindex] = 0;
        } else {
            signal_a[signalindex] = length(signal_vel[signalindex]);
            if (signal_a[signalindex] < MIN_LENGTH) {
                signal_norm[signalindex] = vector(1, 0, 0);
            } else {
                signal_norm[signalindex] = normalize(signal_vel[signalindex]);
            }

            // we have at least one sample per pixel so we can ignore
            // signals that are slow and low frequency
            if (signal_freq[signalindex] < 0.125 && signal_a[signalindex] < 0.25) {
                // mark as no signal
                signal_freq[signalindex] = 0;
                signal_vel[signalindex] = vector(0, 0, 0);
                signal_norm[signalindex] = vector(0, 0, 0);
                signal_a[signalindex] = 0;
            } else {
                avg_vel = avg_vel + signal_vel[signalindex];
                avg_count = avg_count + 1;
            }
        }
    }

    
    if (eff_do_freq_calc == 0 || avg_count == 0) {
        // nothing to calculate, exit
        OutMotion = vector(0, 0, 0);
        OutBounds = vector(0, 0, 0);
    } else {

    // ********************************************************
    // CHECK FOR DIFFERENT SIGNAL DIRECTIONS
    // ********************************************************
        avg_vel = avg_vel / avg_count;
        float avg_a = length(avg_vel);
        vector avg_norm;
        if (avg_a > MIN_LENGTH) {
            avg_norm = normalize(avg_vel);
        } else {
            avg_norm = vector(0, 0, 0);
        }

        // check for signals moving at different velocities
        float diff_velocities = 0;
        for (signalindex = 0; signalindex < num_signals; signalindex += 1) {
            if (signal_freq[signalindex] == 0) {
                // do nothing
            } else {
            
                // signals have frequency content
                if (signal_a[signalindex] < 0.2 || avg_a < 0.2) {
                    // at least one slow moving signal, look at absolute diff
                    if (length(signal_vel[signalindex] - avg_vel) > 0.2) {
                        diff_velocities = 1;
                    } else {
                        diff_velocities = 0;
                    }
                } else {
                    // fast moving signals, look at relative scales
                    // dot product checks if angle is greater than 20 degrees
                    if (((signal_norm[signalindex] .  avg_norm) < MAX_ANGLE_COS) ||
                            (signal_a[signalindex] / avg_a > 1.2) ||
                            (avg_a / signal_a[signalindex] > 1.2)) {
                        diff_velocities = 1;
                    } else {
                        diff_velocities = 0;
                    }
                }
            }
            if (diff_velocities) {
                // break out of loop
                signalindex = 100000;
            }
        }

    // ********************************************************
    // COMPUTE a_min AND a_max
    // ********************************************************
        float a_min = 10000;
        float a_max = -10000;
        for (signalindex = 0; signalindex < num_signals; signalindex += 1) {
            if (signal_freq[signalindex] > 0) {
                a_min = min(a_min, signal_a[signalindex]);
                a_max = max(a_max, signal_a[signalindex]);
            }
        }

        if (a_max == -10000) {
            a_min = 0;
            a_max = 0;
        }

        if (diff_velocities) {
            a_min = 0;
        }

    // ********************************************************
    // COMBINE FREQUENCIES PER LIGHT (SIGNALS MULTIPLY)
    // ********************************************************
        // combine the BRDF/SHADOW/TEXTURE signals per light
        //   BRDF/SHADOW/TEXTURE signals are multiplied per light
        //   so in the fourier domain we need to add signal_freq bounds
        for (lightindex = 0; lightindex < num_lights_wrefl; lightindex += 1) {
            float surfaceindex = SURFACE_INDEX;
            float brdfindex = NONLIGHT_SIGNALS +
                lightindex * LIGHT_SIGNALS + BRDF_OFFSET;
            float shadowindex = NONLIGHT_SIGNALS +
                lightindex * LIGHT_SIGNALS + SHADOW_OFFSET;

            // we store per light results in brdfindex, except for a_max
            // which we store in shadowindex
            signal_freq[brdfindex] = signal_freq[surfaceindex] +
                signal_freq[brdfindex] + signal_freq[shadowindex];
        }


    // ********************************************************
    // COMBINE FREQUENCIES OF ALL LIGHTS (SIGNALS ADD)
    // ********************************************************
        // combine the contribution of each light
        //   per-light signals are all added together for the final color
        //   so in the fourier domain we take the max of signal_freq
        //   frequency bounds
        for (lightindex = 0; lightindex < num_lights_wrefl; lightindex += 1) {
            float brdfindex = NONLIGHT_SIGNALS +
                lightindex * LIGHT_SIGNALS + BRDF_OFFSET;
            signal_freq[0] = max(signal_freq[0], signal_freq[brdfindex]);
        }


    // ********************************************************
    // WRITE OUT FINAL FREQ DATA
    // ********************************************************
        OutMotion = avg_norm;

        OutBounds[0] = a_min;
        OutBounds[1] = a_max;
        OutBounds[2] = signal_freq[0];
    } // if computing frequencies for at least one signal
}



